from check_your_heuristic.dataset.Dataset import Dataset
from check_your_heuristic.heuristics.BasicHeuristics import BasicHeuristics
from check_your_heuristic.utils import get_target_list
from typing import Dict, Any, Union
import pandas as pd


class WordInContextHeuristics(BasicHeuristics):
    """
    Basic Heuristics for the Word-in-Context-like datasets (RUSSE in Russian SuperGLUE)
    """

    def __init__(self, config: Dict[str, Any], dataset: Dataset):
        super(WordInContextHeuristics, self).__init__(dataset=dataset, config=config)
        self.column_1 = config["column_name1"]
        self.column_2 = config["column_name2"]
        self.start1 = config["start1"]
        self.end1 = config["end1"]
        self.start2 = config["start2"]
        self.end2 = config["end2"]
        self.target_list = get_target_list(self.train[self.target_name])

    def heuristic_same_form(self, data: pd.DataFrame, length: int) -> Dict[str, Union[str, Dict[str, str]]]:
        """
        Heuristic that checks whether the words are in the same form
        :param data: data set provided for checking
        :param length: Length of the dataset provided
        :return:
        {
        'coverage': '13.85%',
        'correlation':
        {
        label_1: {False: '83.94%', True: '16.06%'},
        label_2: {False: '87.41%', True: '12.59%'}
        }
        }
        """

        result = {}
        same_form_heuristic = data.apply(
            lambda row:
            row[self.column_1][row[self.start1]:row[self.end1]].lower() ==
            row[self.column_2][row[self.start2]:row[self.end2]].lower(),
            axis=1
        )
        result["coverage"] = f'{same_form_heuristic.sum() / length * 100:.2f}%'
        result["correlation"] = self._get_correlation(
            heuristic_result=same_form_heuristic,
            data=data
        )

        return result

    def check_heuristics(self, render_pandas=False) -> Dict[str, Dict[str, Union[str, Dict[str, Dict[str, str]]]]]:
        """
        Checks how the heuristics are present in the data sets and prints the results
        :return: json-like objects with all the heuristics
        """
        result = dict()

        result["check_substring_train"] = self.check_substring(data=self.train, length=len(self.train))
        result["vocab_intersection_train"] = self.heuristic_vocab_intersection(data=self.train, length=len(self.train))
        result["same_form_heuristic"] = self.heuristic_same_form(data=self.train, length=len(self.train))
        if self.valid is not None:
            result["check_substring_valid"] = self.check_substring(data=self.valid, length=len(self.valid))
            result["vocab_intersection_valid"] = self.heuristic_vocab_intersection(data=self.valid,
                                                                                   length=len(self.valid))
            result["same_form_heuristic"] = self.heuristic_same_form(data=self.train, length=len(self.train))
        if render_pandas:
            result = self._render_pandas_results(res_dict=result)

        return result



from check_your_heuristic.dataset.Dataset import Dataset
from check_your_heuristic.heuristics.Heuristic import BaseHeuristicSolver
from typing import Dict, Any, Union, Tuple
import pandas as pd
from check_your_heuristic.utils import get_target_list
from collections import Counter
import matplotlib.pyplot as plt
import seaborn as sns
import os


class BasicHeuristics(BaseHeuristicSolver):

    def __init__(self, config: Dict[str, Any], dataset: Dataset):
        super(BaseHeuristicSolver, self).__init__(dataset=dataset, config=config)
        self.column_1 = config["column_name1"]
        self.column_2 = config["column_name2"]
        self.target_list = get_target_list(self.train[self.target_name])
        self.output_dir = config["output_dir"] if "output_dir" in config and config["output_dir"] \
            else 'output_check_your_heuristic/'
        if self.output_dir == 'output_check_your_heuristic/':
            os.makedirs('output_check_your_heuristic/', exist_ok=True)

    @staticmethod
    def check_substring_function(text1: str, text2: str) -> Tuple[bool, bool]:
        """
        Heuristic that converts strings to lowercase to
        check whether one is substring of another and vice versa

        :param text1: "The cat is sleeping"
        :param text2: "The cat is sleeping in the box"
        :return: True, False
        """
        text1 = text1.lower()
        text2 = text2.lower()
        left = text1 in text2
        right = text2 in text1

        return left, right

    def check_number_of_words(self, data: pd.DataFrame) -> Dict[str, Dict[str, int]]:
        """
        Function that computes mean and median length of strings in regards to labels
        :param data: data set provided for checking

        Example output
        :return: {
        'mean':
        {'lengths_column1_entailment': 6, 'lengths_column2_entailment': 34,
        'lengths_column1_not_entailment': 6, 'lengths_column2_not_entailment': 32},
        'median':
        {'lengths_column1_entailment': 5, 'lengths_column2_entailment': 32,
        'lengths_column1_not_entailment': 5, 'lengths_column2_not_entailment': 29}}
        """

        result = {"mean": {}, "median": {}}
        data["lengths_column1"] = data[self.column_1].apply(lambda x: len(x.split()))
        data["lengths_column2"] = data[self.column_2].apply(lambda x: len(x.split()))
        for target in self.target_list:
            result["mean"][f"lengths_column1_{target}"] = \
                round(data[data[self.target_name] == target]["lengths_column1"].mean())
            result["mean"][f"lengths_column2_{target}"] = \
                round(data[data[self.target_name] == target]["lengths_column2"].mean())
            result["median"][f"lengths_column1_{target}"] = \
                round(data[data[self.target_name] == target]["lengths_column1"].median())
            result["median"][f"lengths_column2_{target}"] = \
                round(data[data[self.target_name] == target]["lengths_column2"].median())

        self._plot_boxplot(data=data, column_name='lengths_column2', output_name=self.column_2)
        self._plot_boxplot(data=data, column_name='lengths_column1', output_name=self.column_1)
        print(result)
        return result

    def check_substring(self, data: pd.DataFrame, length: int) -> Dict[str, Dict[str, Union[str, Dict[str, str]]]]:
        """
        Check if the heuristic check_substring is in the data frame and returns the percent of rows that have
        this heuristic present as well as the how it correlates with the labels

        :param data: Train or Validation dataset
        :param length: Length of the dataset provided
        :return:  {
        'coverage':
        {
        'hypothesis_premise': '0.92%', 'premise_hypothesis': '0.00%'
        },
        'correlation':
        {
        'hypothesis_premise': {'entailment': {False: '98.68%', True: '1.32%'},
                               'not_entailment': {False: '99.52%', True: '0.48%'}},
        'premise_hypothesis': {'entailment': {False: '100.00%'}, 'not_entailment': {False: '100.00%'}}
        }
        }
        """
        result = {"coverage": {}, "correlation": {}}
        substring_heuristic = pd.DataFrame()
        substring_heuristic[["left", "right"]] = data.apply(
            lambda row: self.check_substring_function(
                text1=row[self.column_1],
                text2=row[self.column_2]
            ),
            axis=1,
            result_type="expand"
        )
        result["coverage"][f"{self.column_1}_{self.column_2}"] = \
            f'{substring_heuristic["left"].sum() / length * 100:.2f}%'
        result["coverage"][f"{self.column_2}_{self.column_1}"] = \
            f'{substring_heuristic["right"].sum() / length:.2f}%'

        result["correlation"][f"{self.column_1}_{self.column_2}"] = self._get_correlation(
            heuristic_result=substring_heuristic["left"],
            data=data
        )
        result["correlation"][f"{self.column_2}_{self.column_1}"] = self._get_correlation(
            heuristic_result=substring_heuristic["right"],
            data=data
        )

        return result

    @staticmethod
    def check_vocab_intersection(text1: str, text2: str) -> Tuple[bool, bool, bool, bool, bool, bool, bool]:
        """
        Heuristic that converts strings to lowercase, splits them into word tokens to
        checks whether their vocabulary intersects by the following thresholds:
        0.1, 0.25, 0.33, 0.5, 0.66, 0.75 and 0.9

        :param text1: "Cat is sleeping in the comfy chair"
        :param text2: "The cat is sitting at the dining table"
        :return: (True, False, False, False, False, False, False)
        """
        tokens1 = set(text1.split())
        tokens2 = set(text2.split())

        ten_percent = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.10
        quarter = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.25
        third = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.33
        half = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.5
        two_thirds = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.66
        two_quarters = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.75
        ninety_percent = len(tokens1 & tokens2) / len(tokens1 | tokens2) > 0.9

        return ten_percent, quarter, third, half, two_thirds, two_quarters, ninety_percent

    def heuristic_vocab_intersection(self, data: pd.DataFrame, length: int) -> \
            Dict[str, Dict[str, Union[str, Dict[str, str]]]]:
        """
        Check if the heuristic check_vocab_intersection is in the data frame and returns the percent of rows that have
        this heuristic present as well as the how it correlates with the labels
        :param data: Train or Validation dataset
        :param length: Length of the dataset provided
        :return:
         {'coverage':
         {
         'threshold_0.1': '23.81%', 'threshold_0.25': '4.01%', 'threshold_0.33': '2.48%', 'threshold_0.5': '0.38%',
         'threshold_0.66': '0.08%', 'threshold_0.75': '0.00%', 'threshold_0.9': '0.00%'
         },
         'correlation':
         {
         'threshold_0.1':
         {'entailment': {True: '25.17%', False: '74.83%'}, 'not_entailment': {True: '22.35%', False: '77.65%'}},
         'threshold_0.25':
         {'entailment': {False: '96.17%', True: '3.83%'}, 'not_entailment': {False: '95.78%', True: '4.22%'}},
         'threshold_0.33':
         {'entailment': {False: '97.87%', True: '2.13%'}, 'not_entailment': {False: '97.14%', True: '2.86%'}},
         'threshold_0.5':
         {'entailment': {False: '99.71%', True: '0.29%'}, 'not_entailment': {False: '99.52%', True: '0.48%'}},
         'threshold_0.66':
         {'entailment': {False: '99.93%', True: '0.07%'}, 'not_entailment': {False: '99.92%', True: '0.08%'}},
         'threshold_0.75':
         {'entailment': {False: '100.00%'}, 'not_entailment': {False: '100.00%'}},
         'threshold_0.9':
        {'entailment': {False: '100.00%'}, 'not_entailment': {False: '100.00%'}}
        }
        }
        """
        result = {"coverage": {}, "correlation": {}}
        thresholds_columns = ["0.1", "0.25", "0.33", "0.5", "0.66", "0.75", "0.9"]
        vocab_intersection = pd.DataFrame(
            columns=thresholds_columns
        )
        vocab_intersection[thresholds_columns] = data.apply(
            lambda row: self.check_vocab_intersection(
                text1=row[self.column_1],
                text2=row[self.column_2]
            ),
            axis=1,
            result_type="expand"
        )
        for element in thresholds_columns:
            result["coverage"][f"threshold_{element}"] = \
                f'{vocab_intersection[element].sum() / length * 100:.2f}%'
            result["correlation"][f"threshold_{element}"] = self._get_correlation(
                heuristic_result=vocab_intersection[element],
                data=data
            )
        return result

    def check_heuristics(self, render_pandas=False) -> Dict[str, Dict[str, Union[str, Dict[str, Dict[str, str]]]]]:
        """
        Checks how the heuristics are present in the data sets and prints the results
        :return: json-like objects with all the heuristics
        """
        result = dict()
        result["check_substring_train"] = self.check_substring(data=self.train, length=len(self.train))
        result["vocab_intersection_train"] = self.heuristic_vocab_intersection(data=self.train, length=len(self.train))
        if self.valid is not None:
            result["check_substring_valid"] = self.check_substring(data=self.valid, length=len(self.valid))
            result["vocab_intersection_valid"] = self.heuristic_vocab_intersection(data=self.valid,
                                                                                   length=len(self.valid))
        if render_pandas:
            result = self._render_pandas_results(res_dict=result)

        return result

    def _get_correlation(self, heuristic_result: pd.Series, data: pd.DataFrame) -> Dict[str, str]:
        """
        Computes the correlation between the heuristic results and the labels
        labels are taken from the dataset provided for checking

        :param data: dataset provided for checking
        :param heuristic_result: pd.Series([1, 2, 2, 1])
        :return: {'label_1': {1: '50.00%'}, 'label_2': {2: '50.00%'}}
        """
        correlation = {}
        for target in self.target_list:
            indexes = data[data[self.target_name] == target].index
            samples = heuristic_result.iloc[indexes].to_list()
            counts = dict(Counter(samples))
            counts = {k: f"{v / len(samples) * 100:.2f}%" for k, v in counts.items()}
            correlation[target] = counts
        return correlation

    def _render_pandas_results(
            self,
            res_dict: Dict[str, Dict[str, Union[str, Dict[str, Dict[str, str]]]]]
    ) -> pd.DataFrame:
        columns = ['heuristic', "additional_info", 'coverage']
        for target in self.target_list:
            columns.append(f'correlation_{target}')

        df = pd.DataFrame(columns=columns)

        for k in res_dict.keys():
            for key in res_dict[k]['coverage'].keys():
                res = {"heuristic": k, "additional_info": key, "coverage": res_dict[k]['coverage'][key]}
                for label in self.target_list:
                    r = res_dict[k]["correlation"][key][label]
                    if True in r:
                        res[f"correlation_{label}"] = res_dict[k]["correlation"][key][label][True]
                    else:
                        res[f"correlation_{label}"] = "0.00%"
                df = df.append(res, ignore_index=True)
        return df

    def _plot_boxplot(self, data: pd.DataFrame, column_name: str, output_name=str) -> None:
        """

        :param data: data frame provided for checking
        :param column_name: column to compute box plot
        :param output_name: file_name to save
        :return: None
        """
        sns.boxplot(x=self.target_name, y=column_name, data=data)
        plt.xlabel("Labels")
        plt.ylabel("Number of words")
        plt.title(f'Relation between label and number of words in {output_name}', fontsize=14)
        plt.savefig(f"{self.output_dir}/lengths_{output_name}.png")
        plt.close()

    def get_visuals(self):

        _ = plt.title('Label distribution in train data', fontsize=14)
        _ = plt.pie(self.train[self.target_name].value_counts(), autopct="%.1f%%", explode=[0.05] * 2,
                    labels=self.train[self.target_name].value_counts().keys(), pctdistance=0.5, textprops=dict(fontsize=12))
        plt.savefig(f"{self.output_dir}/Label_distribution_in_train_data.png")
        plt.close()
        self.check_number_of_words(data=self.train)

        if self.valid is not None:
            _ = plt.title('Label distribution in validation data', fontsize=14)
            _ = plt.pie(self.valid[self.target_name].value_counts(), autopct="%.1f%%", explode=[0.05] * 2,
                        labels=self.valid[self.target_name].value_counts().keys(), pctdistance=0.5,
                        textprops=dict(fontsize=12))
            plt.savefig(f"{self.output_dir}/Label_distribution_in_validation_data.png")
            plt.close()
            self.check_number_of_words(data=self.valid)




from check_your_heuristic.utils import load_config, get_argparse
from check_your_heuristic.dataset import Dataset, MultiRCDataset, ReCoRDDataset
from check_your_heuristic.heuristics import BasicHeuristics, ReCoRDHeuristics, WordInContextHeuristics


def base_case():
    parser = get_argparse()
    args = parser.parse_args()

    config = load_config(args.path_to_config)

    dataset = Dataset(path=config['train_dataset_dir'], path_valid=config['valid_dataset_dir'])
    solver = BasicHeuristics(dataset=dataset, config=config)

    solver.get_visuals()
    solver.check_heuristics()
    solver.all_methods()


def multirc_case():
    parser = get_argparse()
    args = parser.parse_args()

    config = load_config(args.path_to_config)

    dataset = MultiRCDataset(path=config['train_dataset_dir'], path_valid=config['valid_dataset_dir'])
    solver = BasicHeuristics(dataset=dataset, config=config)

    solver.get_visuals()
    solver.check_heuristics()
    solver.all_methods()


def record_case():
    parser = get_argparse()
    args = parser.parse_args()

    config = load_config(args.path_to_config)

    dataset = ReCoRDDataset(path=config['train_dataset_dir'])
    solver = ReCoRDHeuristics(dataset=dataset, config=config)

    solver.check_heuristics()


def wordincontext_case():
    parser = get_argparse()
    args = parser.parse_args()

    config = load_config(args.path_to_config)

    dataset = Dataset(path=config['train_dataset_dir'])
    solver = WordInContextHeuristics(dataset=dataset, config=config)

    solver.get_visuals()
    solver.check_heuristics()
    solver.all_methods()




import os
import random
import numpy as np
from sklearn.metrics import classification_report
import pandas as pd
from nptyping import NDArray
from typing import Dict, Any, List, Union
from check_your_heuristic.dataset import Dataset
import warnings

warnings.filterwarnings("ignore")


def seed_everything(seed: int):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.RandomState(seed)
    np.random.seed(seed)


seed_everything(42)


class BaseSolver:

    def __init__(self, config: Dict[str, Any], dataset: Dataset, seed: int = 42):  # TODO add typehint

        self.train = dataset.train
        self.valid = dataset.valid
        self.seed = seed

        self.target_name = config["target_name"]
    
    def all_methods(self):

        if self.valid is not None:
            test_size = len(self.valid)
            y_true = list(self.valid.label)
        else:
            print("There are no Validation/Test set in this task")
            print("Making Predictions for Train dataset")
            test_size = len(self.train)
            y_true = self.train[self.target_name]
            
        print()
        print(f"Making Prediction based on Majority Class")
        y_pred = self.majority_class(test_size=test_size)
        self.show_report(y_true, y_pred)

        print()
        print(f"Making Prediction based on Random Choice")
        y_pred = self.random_choice(test_size=test_size)
        self.show_report(y_true, y_pred)
        
        print()
        print(f"Making Prediction based on Random Choice Considered Classes Distribution")
        y_pred = self.random_balanced_choice(test_size=test_size)
        self.show_report(y_true, y_pred)

    def show_report(self, y_true: pd.Series, y_pred: pd.Series):
        print(classification_report(y_true, y_pred))

    def majority_class(self, test_size: int) -> List[Union[str, int]]:
        """
        Make prediction based on majority class of train dataset
        test_size: how many predictions should be made
        return: List of predictions
        """

        prediction = self.train[self.target_name].mode()[0]
        y_pred = [prediction] * test_size
        return y_pred

    def random_choice(self, test_size: int) -> NDArray[Any]:
        """
        Make random predictions
        label: label column in df (str)
        test_size: how many predictions should be made
        return: List of predictions
        """
        options = sorted(self.train[self.target_name].unique())
        if test_size != 1:
            np.random.seed(self.seed)
        y_pred = np.random.choice(options, size=test_size)
        return y_pred

    def random_balanced_choice(self, test_size: int) -> NDArray[Any]:
        """
        Make random predictions with calculated probabilities
        label: label column in df (str)
        test_size: how many predictions should be made
        return: List of predictions
        """
        frequencies = dict(self.train[self.target_name].value_counts(normalize=True))

        labels = []
        probs = []
        for key, value in frequencies.items():
            labels.append(key)
            probs.append(value)
        if test_size != 1:
            np.random.seed(self.seed)
        y_pred = np.random.choice(labels, size=test_size, p=probs)
        return y_pred




import unittest

from parameterized import parameterized

from check_your_heuristic.Base import BaseSolver
from check_your_heuristic.dataset.Dataset import Dataset
from check_your_heuristic.utils import load_config


class TestBaseSolver(unittest.TestCase):
    """Class for testing BaseSolver."""

    @classmethod
    def setUpClass(cls) -> None:
        """SetUp tests with lemmatizer and preprocessor."""

        config = load_config(path="configs/config.yaml")
        dataset = Dataset(
            path="data/train.jsonl",
            path_valid="data/val.jsonl",
            path_test="data/test.jsonl"
        )
        solver = BaseSolver(
            config=config,
            dataset=dataset,
            seed=23
        )
        cls.dataset = dataset
        cls.solver = solver 

    @parameterized.expand(
        [
            (1, ["entailment"]),
            (3, ["entailment", "entailment", "entailment"]),
        ]
    )
    def test_majority_class(self, size, labels_true) -> None:
        """Testing majority class prediction"""
        labels_pred = self.solver.majority_class(test_size=size)
        self.assertListEqual(labels_true, labels_pred)

    @parameterized.expand(
        [
            (1, ["not_entailment"]),
            (3, ["not_entailment", "entailment", "entailment"]),
        ]
    )
    def test_random_choice(self, size, labels_true) -> None:
        """Testing random choice prediction"""

        labels_pred = self.solver.random_choice(test_size=size)
        labels_pred = list(labels_pred)
        self.assertListEqual(labels_true, labels_pred)
    
    @parameterized.expand(
        [
            (1, ["entailment"]),
            (3, ["entailment", "not_entailment", "not_entailment"]),
        ]
    )
    def test_random_balanced_choice(self, size, labels_true) -> None:
        """Testing random balanced choice prediction"""

        labels_pred = self.solver.random_balanced_choice(test_size=size)
        labels_pred = list(labels_pred)
        self.assertListEqual(labels_true, labels_pred)



import pandas as pd
from pathlib import Path
from check_your_heuristic.dataset.MultiRCDataset import MultiRCDataset
from typing import Any, Tuple


class ReCoRDDataset(MultiRCDataset):
    """
    ReCorD is Reading Comprehension with Commonsense Reasoning dataset in SuperGLUE benchmark.
    The corresponding name for this type of datasets in Russian SuperGLUE is RuCoS or
    Russian reading comprehension with Commonsense reasoning.

    NB! As of now this class can only work with json data format
    """

    def __init__(self, path: str, path_valid: str = None, path_test: str = None):
        super(ReCoRDDataset, self).__init__(path=path, path_valid=path_valid, path_test=path_test)

    def split_texts_and_questions(self, line) -> Tuple[str, Any, Any]:
        """ transforms a complex json object into a single row dataframe"""
        text = line['passage']['text']
        entities = line['passage']['entities']
        questions = line['qas']

        return text, entities, questions

    def read_data(self, path: Path) -> pd.DataFrame:
        """ get json file content as a pandas DataFrame"""
        df = pd.DataFrame()

        lines = self.yield_lines(path=path)

        for passage_id, line in enumerate(lines):
            text, entities, questions = self.split_texts_and_questions(line=line)

            for i in range(len(questions)):
                df = df.append(
                    {
                        'text': text,
                        'entities': entities,
                        'question': questions[i]['query'],
                        'answers': questions[i]['answers']
                    },
                    ignore_index=True)
        return df




import pandas as pd
from pathlib import Path
from check_your_heuristic.dataset.BaseDataset import BaseDataset
import os


class Dataset(BaseDataset):
    """
    Dataset class for datasets that have simple structure
    """
    def __init__(self, path: str, path_valid: str = None, path_test: str = None):

        self.valid_data_types = {
            '.csv': self._read_csv,
            '.xls': self._read_excel,
            '.xlsx': self._read_excel,
            '.json': self._read_json,
            '.jsonl': self._read_json
        }

        super(Dataset, self).__init__(path=path, path_valid=path_valid, path_test=path_test)

    def read_data(self, path: Path) -> pd.DataFrame:
        """
        Given the path to the file returns extension of that file

        Example:
        path: "../input/some_data.csv"
        :return: ".csv"
        """
        _, extension = os.path.splitext(path)
        if extension.lower() in self.valid_data_types:
            return self.valid_data_types[extension](path)
        else:
            raise ValueError(f"Your data type ({extension}) is not supported, please convert your dataset "
                             f"to one of the following formats {list(self.valid_data_types.keys())}.")

    @staticmethod
    def _read_csv(path: Path) -> pd.DataFrame:
        """
        Reads a csv file given its path
        :param path: Path("../../some_file.csv")
        :return: dataframe
        """
        return pd.read_csv(filepath_or_buffer=path, encoding="utf-8")

    @staticmethod
    def _read_excel(path: Path) -> pd.DataFrame:
        """
        Reads a xls or xlsx file given its path
        :param path: Path("../../some_file.xlsx")
        :return: dataframe
        """
        return pd.read_excel(io=path, engine="openpyxl", encoding="utf-8")

    @staticmethod
    def _read_json(path: Path) -> pd.DataFrame:
        """
        Reads a json or jsonl file given its path
        :param path: Path("../../some_file.jsonl")
        :return: dataframe
        """
        return pd.read_json(path_or_buf=path, lines=True, encoding="utf-8")






import pandas as pd
from pathlib import Path
from check_your_heuristic.dataset.BaseDataset import BaseDataset
from typing import Iterator, Tuple, Any
import json
import logging


class MultiRCDataset(BaseDataset):
    """
    MultiRC is Multi-Sentence Reading Comprehension dataset in SuperGLUE benchmark.
    The corresponding name for this type of datasets in Russian SuperGLUE is MuSeRC or
    Russian Multi-Sentence Reading Comprehension.

    NB! As of now this class can only work with json data format
    """

    def __init__(self, path: str, path_valid: str = None, path_test: str = None):
        logging.warning("As of now this class can only work with json data format.")
        super(MultiRCDataset, self).__init__(path=path, path_valid=path_valid, path_test=path_test)

    def read_data(self, path: Path) -> pd.DataFrame:
        """ get json file content as a pandas DataFrame"""

        df = pd.DataFrame(columns=['question', 'text', 'label', 'passage'])

        lines = self.yield_lines(path=path)

        for passage_id, line in enumerate(lines):
            passage, questions = self.split_texts_and_questions(line=line)
            questions = pd.json_normalize(
                questions, 'answers', 'question'
            )[['question', 'text', 'label']]

            questions['passage'] = passage

            df = pd.concat([df, questions])

        df = df.rename(columns={'text': 'answer'})

        return df

    @staticmethod
    def yield_lines(path: Path) -> Iterator:
        """ yields json lines one by one """
        with open(path, encoding="utf-8") as f:
            for line in f:
                yield json.loads(line)

    def split_texts_and_questions(self, line) -> Tuple[str, Any]:
        """ transforms a complex json object into a single row dataframe"""
        text = line['passage']['text']
        questions = line['passage']['questions']
        return text, questions



from check_your_heuristic.dataset.ReCoRDDataset import ReCoRDDataset
from check_your_heuristic.heuristics.Heuristic import BaseHeuristicSolver
from typing import Dict, Any, List
import pandas as pd
import string
import numpy as np
import logging


class ReCoRDHeuristics(BaseHeuristicSolver):
    def __init__(self, config: Dict[str, Any], dataset: ReCoRDDataset):
        super(BaseHeuristicSolver, self).__init__(dataset=dataset, config=config)
        self.passage_column = config["passage_column"]
        self.question_column = config["question_column"]
        self.entities_column = config["entities_column"]

    @staticmethod
    def normalize_answer(text: str):
        """Lower text and remove punctuation, articles and extra whitespace."""

        def white_space_fix(line):
            return ' '.join(line.split())

        def remove_punct(line):
            exclude = set(string.punctuation)
            return ''.join(ch for ch in line if ch not in exclude)

        return white_space_fix(remove_punct(text.lower()))

    def _get_entities(self, row, column_name:str):
        words = [
            row[self.passage_column][x["start"]: x["end"]]
            for x in row[column_name]
        ]
        return words

    def get_basic_pred(self,
                       row: pd.DataFrame,
                       words: List[str],
                       _words: List[str],
                       line_candidates: List[str]
                       ) -> str:
        if len(_words) == 0:
            if len(words) == 1:
                pred = words[0]
            else:
                for word in words:
                    line_candidates.append(row[self.question_column].replace("@placeholder", word))
                pred_idx = np.random.choice(np.arange(1, len(line_candidates)),
                                            size=1)[0]
                pred = np.array(words)[pred_idx]
        elif len(_words) == 1:
            pred = _words[0]
        else:
            for word in _words:
                line_candidates.append(row[self.question_column].replace("@placeholder", word))
            pred_idx = np.random.choice(np.arange(1, len(line_candidates)),
                                        size=1)[0]
            pred = np.array(_words)[pred_idx]
        return pred

    def filtration_count_heuristic(self, row: pd.DataFrame) -> str:
        """
        Heuristic that removes some candidates and filters out candidates depended on times they occurred in the text
        If there are a lot of candidates the one is chosen randomly
        """
        line_candidates = []
        _words = []
        text = row[self.passage_column].split()
        words = self._get_entities(row=row, column_name=self.entities_column)
        for word in words:
            if word[:-2] not in row[self.question_column] or text.count(words[:-2]) >= 2:
                _words.append(word)
        pred = self.get_basic_pred(row=row, words=words, _words=_words, line_candidates=line_candidates)
        return self.normalize_answer(pred)

    def remove_candidates_heuristic(self, row: pd.DataFrame) -> str:
        """
        Heuristic that removes candidates that occur in the question
        """
        words = self._get_entities(row=row, column_name=self.entities_column)
        line_candidates = []
        _words = []
        for word in words:
            if word[:-1] not in row[self.question_column]:
                _words.append(word)

        pred = self.get_basic_pred(row=row, words=words, _words=_words, line_candidates=line_candidates)
        return self.normalize_answer(pred)

    def metric_max_over_ground_truths(self, row: pd.DataFrame, predictions_colname: str) -> float:
        """
        As there is several true answers, we go over all and compute metric for all
        :param row: row of the data frame predicted
        :param predictions_colname: the name for heuristic
        :return:
        """
        scores_for_ground_truths = [0]
        prediction = row[predictions_colname]
        ground_truths = self._get_entities(row=row, column_name=self.target_name)
        for ground_truth in ground_truths:
            score = self.exact_match_score(prediction, ground_truth)
            scores_for_ground_truths.append(score)
        return max(scores_for_ground_truths)

    def check_heuristics(self) -> Dict[str, float]:
        """
        Checks how the heuristics are present in the data sets and prints the results
        :return: json-like object with all the results
        """

        result = {}

        self.train["pred_remove_candidates_heuristic"] = self.train.apply(
            self.remove_candidates_heuristic,
            axis=1
        )
        result["exact_match_score_remove_candidates_heuristic"] = np.mean(self.train.apply(
            lambda row: self.metric_max_over_ground_truths(
                row=row,
                predictions_colname="pred_remove_candidates_heuristic"
            ),
            axis=1).to_list()
        )

        self.train["pred_filtration_count_heuristic"] = self.train.apply(
            self.filtration_count_heuristic,
            axis=1
        )
        self.train["true_filtration_count_heuristic"] = self.train[self.target_name]
        result["exact_match_score_filtration_count_heuristic_train"] = np.mean(self.train.apply(
            lambda row: self.metric_max_over_ground_truths(
                row=row,
                predictions_colname="true_filtration_count_heuristic"
            ),
            axis=1
        ).to_list()
        )
        if self.valid is not None:
            result_df = pd.DataFrame()
            result_df["pred_remove_candidates_heuristic"] = self.valid.apply(
                self.remove_candidates_heuristic,
                axis=1
            )
            result_df["true_labels"] = self.valid[self.target_name]
            result["exact_match_score_remove_candidates_heuristic_valid"] = np.mean(result_df.apply(
                lambda row: self.metric_max_over_ground_truths(
                    row=row,
                    predictions_colname="pred_remove_candidates_heuristic"
                ),
                axis=1
                ).to_list()
            )

            result_df["pred_filtration_count_heuristic"] = self.valid.apply(
                self.filtration_count_heuristic,
                axis=1
            )
            result_df["true_filtration_count_heuristic"] = self.valid[self.target_name]
            result["exact_match_score_filtration_count_heuristic_valid"] = np.mean(result_df.apply(
                lambda row: self.metric_max_over_ground_truths(
                    row=row,
                    predictions_colname="pred_filtration_count_heuristic"
                ),
                axis=1).to_list()
            )
        for key, value in result.items():
            print(key, '\n', value, '\n')

        return result

    def exact_match_score(self, prediction: str, ground_truth: str) -> bool:
        return prediction == self.normalize_answer(ground_truth)

    def all_methods(self):
        logging.error("Method is deprecated for this type of dataset")
        raise AttributeError

    def random_balanced_choice(self):
        logging.error("Method is deprecated for this type of dataset")
        raise AttributeError

    def random_choice(self):
        logging.error("Method is deprecated for this type of dataset")
        raise AttributeError

    def majority_class(self):
        logging.error("Method is deprecated for this type of dataset")
        raise AttributeError

    def show_report(self):
        logging.error("Method is deprecated for this type of dataset")
        raise AttributeError







from .BasicHeuristics import BasicHeuristics
from .ReCoRDHeuristics import ReCoRDHeuristics
from .WordInContextHeuristics import WordInContextHeuristics

__all__ = [
    "BasicHeuristics",
    "ReCoRDHeuristics",
    "WordInContextHeuristics",
]


base_config = dict(
    train_dataset_dir="",
    valid_dataset_dir="",
    column_name1="hypothesis",
    column_name2="premise",
    target_name="label",
)

wordincontext_config = dict(
    train_dataset_dir="",
    valid_dataset_dir="",
    column_name1="sentence1",
    column_name2="sentence2",
    start1="start1",
    start2="start2",
    end1="end1",
    end2="end2",
    target_name="label",
)

record_config = dict(
        train_dataset_dir="",
        passage_column='text',
        question_column="question",
        entities_column="entities",
        target_name="answers",
    )


from check_your_heuristic.Base import BaseSolver
from typing import Dict, Any
from check_your_heuristic.dataset.Dataset import Dataset


class BaseHeuristicSolver(BaseSolver):
   
    def __init__(self, config: Dict[str, Any], dataset: Dataset):
        super(BaseHeuristicSolver, self).__init__(dataset=dataset, config=config)
    
    def check_heuristics(self):
        raise NotImplementedError



import pandas as pd
from pathlib import Path


class BaseDataset:

    def __init__(self, path: str, path_valid: str = None, path_test: str = None):
        self.train = self.read_data(Path(repr(path)[1:-1]).as_posix())
        self.valid = self.read_data(Path(repr(path_valid)[1:-1]).as_posix()) if path_valid is not None else None
        self.test = self.read_data(Path(repr(path_test)[1:-1]).as_posix()) if path_test is not None else None

    def read_data(self, path: Path) -> pd.DataFrame:
        raise NotImplementedError






from setuptools import setup, find_packages

setup(
    name='check_your_heuristic',
    version='0.0.1',
    license='MIT',
    description='Small library which aim is to check your dataset for being solved by simple heuristics. Multilingual!',
    author='Tatiana Iazykova, Olga Bystrova, Denis Kapelyushnik',
    author_email='tania_yazykova@bk.ru',
    keywords=['heuristics', 'rule-based', 'language models', 'natural language understanding', 'nlp'],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'run-base-case = check_your_heuristic.main:base_case',
            'run-multirc-case = check_your_heuristic.main:multirc_case',
            'run-record-case = check_your_heuristic.main:record_case',
            'run-wordincontext-case = check_your_heuristic.main:wordincontext_case',
        ],
    },
    install_requires=[
          'numpy>=1.19.0',
          'pandas>=1.1.5',
          'scikit-learn>=0.24.2',
          'pymorphy2>=0.9.1',
          'pyyaml',
          'nptyping>=1.4.4',
          'parameterized>=0.8.1',
          'openpyxl>=3.0.9',
          'matplotlib>=3.4.3',
          'seaborn>=0.11.0'
      ],
    classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    ],
    )



import pandas as pd
from typing import List
import yaml
from argparse import ArgumentParser


def get_target_list(target_column: pd.Series) -> List:
    return target_column.unique().tolist()


def load_config(path: str):
    with open(path, "r") as yamlfile:
        data = yaml.safe_load(yamlfile)
        return data


def get_argparse() -> ArgumentParser:
    """Get argument parser.
    Returns:
        ArgumentParser: Argument parser.
    """

    parser = ArgumentParser()
    parser.add_argument(
        "--path_to_config",
        type=str,
        required=True,
        help="Path to config",
    )

    return parser